type RedefineValueTypes<ShapeToRedefine, NewValueType> = {
  [K in keyof ShapeToRedefine]: NewValueType;
};

/** Types that define the data type of a column in the DB */
export type DataType = "TEXT" | "INTEGER" | "REAL" | "BLOB" | "NULL";

export interface Column {
  /** The data type to store this column */
  type: DataType;
  nullable: boolean;
  primary: boolean;

  /** Any additional freeform SQL */
  misc: string;

  /** If set, creates a foreign key association. See Table.reference() */
  references: string;
  cascade: boolean;
  default: string | number | undefined;
  unique: boolean;
  autoincrement: boolean;
}

export type ColumnShorthand = DataType | Partial<Column>;

export type ColumnShorthandMap<T> = RedefineValueTypes<T, ColumnShorthand>;
export type ColumnMap<T> = RedefineValueTypes<T, Column>;

export function generateColumnDefinitionSQL(
  name: string,
  colDef: Column,
): string {
  const parts: string[] = [`"${name}"`, colDef.type];

  if (colDef.primary) {
    parts.push("PRIMARY KEY");
  }

  if (!colDef.primary && !colDef.nullable) {
    parts.push("NOT NULL");
  }

  if (colDef.unique) {
    parts.push("UNIQUE");
  }

  if (colDef.autoincrement) {
    parts.push("AUTOINCREMENT");
  }

  if (colDef.default !== undefined) {
    parts.push(`DEFAULT (${colDef.default})`);
  }

  if (colDef.references.length) {
    parts.push(`REFERENCES ${colDef.references}`);

    if (colDef.cascade) {
      parts.push("ON DELETE CASCADE");
    }
  }

  if (colDef.misc.length) {
    parts.push(colDef.misc);
  }

  return parts.join(" ");
}

function expandColumnShorthand(col: ColumnShorthand): Column {
  if (typeof col === "string") {
    return {
      type: col,
      nullable: false,
      primary: false,
      misc: "",
      references: "",
      cascade: false,
      default: undefined,
      unique: false,
      autoincrement: false,
    };
  }

  return {
    type: col.type ?? "TEXT",
    nullable: col.nullable ?? false,
    primary: col.primary ?? false,
    misc: col.misc ?? "",
    references: col.references ?? "",
    cascade: col.cascade ?? false,
    default: col.default ?? undefined,
    unique: col.unique ?? false,
    autoincrement: col.autoincrement ?? false,
  };
}

export function normalizeColumns<T>(cols: ColumnShorthandMap<T>): ColumnMap<T> {
  const keys = Object.keys(cols);

  const retval: Record<string, Column> = {};

  for (const key of keys) {
    const val = cols[key as keyof typeof cols];

    if (val === undefined) {
      throw new Error("Invariant: Undefined column descriptor");
    }

    retval[key] = expandColumnShorthand(val);
  }

  return retval as ColumnMap<T>;
}
